# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================

package KL::AppBase;

use strict;
use KL::AppInfo;
use KL::Utils;
use File::Basename;
use Getopt::Long;
use Data::Dumper;
use POSIX qw(strftime);

my $ENV_AUTO_INSTALL_FILE = 'KLI_AUTO_INSTALL_FILE';
my $ENV_AUTO_INSTALL_SAVE = 'KLI_AUTO_INSTALL_SAVE';


my $MSG_WARNING_OBSOLETED_ENV_SAVEFILE=<<EOT;
Using of environment variable $ENV_AUTO_INSTALL_SAVE for saving answers is obsoletted. In next versions only --create-auto-install command line option will be supported.
EOT

my $MSG_WARNING_OBSOLETED_ENV_FILE=<<EOT;
Using of environment variable $ENV_AUTO_INSTALL_FILE for auto-install is obsoleted. In next versions only --auto-install command line option will be supported.
EOT

my $_HEADER_SHOW_EULA = <<EOT;
Accepting the End User License Agreement (EULA)
EOT

my $_MSG_SHOW_EULA_BEGIN = <<EOT;
Please read and accept the End User License Agreement (EULA) to continue.
EOT

my $_MSG_SHOW_EULA_HOWTO_QUIT = <<EOT;

<B>NOTE:</B> To quit the EULA viewer, press the "q" key.
EOT

my $_MSG_SHOW_EULA_END = <<EOT;

To display the EULA, press Enter
EOT

my $_MSG_SHOW_EULA_QUESTION = <<EOT;
Do you accept the EULA? (yes/no)
EOT

my $_MSG_SHOW_EULA_NOT_ACCEPTED = <<EOT;
EULA was not accepted.
EOT

my $_MSG_SHOW_EULA_ERROR = <<EOT;
Couldn\'t display the EULA file %1: %2.
EOT

my $MSGS_EULA = {
    'header'   => $_HEADER_SHOW_EULA,
    'begin'    => $_MSG_SHOW_EULA_BEGIN,
    'end'      => $_MSG_SHOW_EULA_END,
    'quit'     => $_MSG_SHOW_EULA_HOWTO_QUIT,
    'question' => $_MSG_SHOW_EULA_QUESTION,
    'declined' => $_MSG_SHOW_EULA_NOT_ACCEPTED,
    'error'    => $_MSG_SHOW_EULA_ERROR,
};

my $_HEADER_KSN_SHOW_EULA = <<EOT;
Reading the Kaspersky Security Network (KSN) Statement
EOT

my $_MSG_KSN_SHOW_EULA_BEGIN  = <<EOT;
Please read the Kaspersky Security Network (KSN) Statement before agreeing or refusing to participate in Kaspersky Security Network.
EOT

my $_MSG_KSN_SHOW_EULA_HOWTO_QUIT = <<EOT;

<B>NOTE:</B> To quit the KSN Statement viewer, press the \"q\" key.
EOT

my $_MSG_KSN_SHOW_EULA_END = <<EOT;

To display the KSN Statement, press Enter
EOT

my $_MSG_KSN_SHOW_EULA_ERROR = <<EOT;
Couldn\'t display the KSN Statement file %1: %2.
EOT

my $_MSG_KSN_SHOW_EULA_QUESTION = <<EOT;
Do you agree to participate in Kaspersky Security Network? (yes/no)
EOT

my $_MSG_KSN_SHOW_EULA_NOT_ACCEPTED = <<EOT;
Kaspersky Security Network participation was declined. KSN features are disabled.

EOT

my $MSGS_EULA_KSN = {
    'header'   => $_HEADER_KSN_SHOW_EULA,
    'begin'    => $_MSG_KSN_SHOW_EULA_BEGIN,
    'end'      => $_MSG_KSN_SHOW_EULA_END,
    'quit'     => $_MSG_KSN_SHOW_EULA_HOWTO_QUIT,
    'question' => $_MSG_KSN_SHOW_EULA_QUESTION,
    'declined' => $_MSG_KSN_SHOW_EULA_NOT_ACCEPTED,
    'error'    => $_MSG_KSN_SHOW_EULA_ERROR,
};

my $_HEADER_KSN_STATISTICS_SHOW_EULA = <<EOT;
Reading the Additional Kaspersky Security Network (KSN) Statement on sending statistics to Kaspersky Lab.
EOT

my $_MSG_KSN_STATISTICS_SHOW_EULA_BEGIN = <<EOT;
Please read the Additional Kaspersky Security Network (KSN) Statement before agreeing or refusing to send statistics.
EOT

my $_MSG_KSN_STATISTICS_SHOW_EULA_HOWTO_QUIT = <<EOT;

<B>NOTE:</B> To quit the Additional KSN Statement viewer, press the \"q\" key.
EOT

my $_MSG_KSN_STATISTICS_SHOW_EULA_END = <<EOT;

To display the Additional KSN Statement, press Enter
EOT

my $_MSG_KSN_STATISTICS_SHOW_EULA_ERROR = <<EOT;
Unable to display the KSN Statement file %1: %2.
EOT

my $_MSG_KSN_STATISTICS_SHOW_EULA_QUESTION = <<EOT;
Do you agree to send statistics to Kaspersky Security Network? (yes/no)
EOT

my $_MSG_KSN_STATISTICS_SHOW_EULA_NOT_ACCEPTED = <<EOT;
You refused to send statistics to Kaspersky Security Network.
EOT

my $MSGS_EULA_KSN_STATISTICS = {
    'header'   => $_HEADER_KSN_STATISTICS_SHOW_EULA,
    'begin'    => $_MSG_KSN_STATISTICS_SHOW_EULA_BEGIN,
    'end'      => $_MSG_KSN_STATISTICS_SHOW_EULA_END,
    'quit'     => $_MSG_KSN_STATISTICS_SHOW_EULA_HOWTO_QUIT,
    'question' => $_MSG_KSN_STATISTICS_SHOW_EULA_QUESTION,
    'declined' => $_MSG_KSN_STATISTICS_SHOW_EULA_NOT_ACCEPTED,
    'error'    => $_MSG_KSN_STATISTICS_SHOW_EULA_ERROR,
};

my $_HEADER_SHOW_PRIVACY_POLICY = <<EOT;
Accepting the Privacy Policy
EOT

my $_MSG_SHOW_PRIVACY_POLICY_BEGIN = <<EOT;
Please read and accept the Privacy Policy to continue.
EOT

my $_MSG_SHOW_PRIVACY_POLICY_HOWTO_QUIT = <<EOT;

<B>NOTE:</B> To quit the Privacy Policy viewer, press the "q" key.
EOT

my $_MSG_SHOW_PRIVACY_POLICY_END = <<EOT;

To display the Privacy Policy, press Enter
EOT

my $_MSG_SHOW_PRIVACY_POLICY_QUESTION = <<EOT;
Do you accept the Privacy Policy? (yes/no)
EOT

my $_MSG_SHOW_PRIVACY_POLICY_NOT_ACCEPTED = <<EOT;
Privacy Policy was not accepted.
EOT

my $_MSG_SHOW_PRIVACY_POLICY_ERROR = <<EOT;
Couldn\'t display the Privacy Policy file %1: %2.
EOT

my $MSGS_PRIVACY_POLICY = {
    'header'   => $_HEADER_SHOW_PRIVACY_POLICY,
    'begin'    => $_MSG_SHOW_PRIVACY_POLICY_BEGIN,
    'end'      => $_MSG_SHOW_PRIVACY_POLICY_END,
    'quit'     => $_MSG_SHOW_PRIVACY_POLICY_HOWTO_QUIT,
    'question' => $_MSG_SHOW_PRIVACY_POLICY_QUESTION,
    'declined' => $_MSG_SHOW_PRIVACY_POLICY_NOT_ACCEPTED,
    'error'    => $_MSG_SHOW_PRIVACY_POLICY_ERROR,
};

my $MSG_FAIL_OPENDIR = <<EOT;
Can\'t open directory %1
EOT

my $MSG_NO_LICENSE = <<EOT;
No license found. Please reinstall %1
EOT

my $MSG_SELECT_LANG = <<EOT;
Please select the language in which you want to read EULA and KSN agreement:
%1
Please select language
EOT

my $HEADER_INSTALL_KEY = <<EOT;
Installing the license
EOT

my $MSG_INSTALL_KEY_ENTER_PATH = <<EOT;
The key file (a file with the .key extension) contains information about your license. You need to install it to use the application.
To install it now, enter the path to your key file (or enter an empty string to continue without installing the key file):
EOT

my $MSG_INSTALL_KEY_ENTER_OTHER_PATH = <<EOT;

Enter another path to your key file (or enter an empty string to continue without installing the key file):
EOT

my $MSG_INSTALL_OPENDIR_FAILED = <<EOT;
Could not open the %1 directory: %2.
EOT

my $MSG_INSTALL_KEY_SUCCESS = <<EOT;
The license from %1 has been installed.
EOT

my $MSG_INSTALL_KEY_FAILED = <<EOT;
The license from %1 could not be installed: %2
EOT

my $MSG_INSTALL_KEY_REJECTED = <<EOT;
The %1 key file was rejected because it is not suitable for this application: %2
EOT

my $MSG_INSTALL_KEY_INVALID_PATH = <<EOT;
The specified path is incorrect.
EOT

my $MSG_INSTALL_KEY_NO_VALID_FOUND = <<EOT;
None of the found key files is correct for this application.
EOT

my $MSG_INSTALL_KEY_NO_KEYS_FOUND = <<EOT;
No key files (files with the .key extension) were found in the %1 directory.
EOT

my $HEADER_SETUP_PROXY = <<EOT;
Configuring the proxy settings to connect to the updates source
EOT

my $MSG_SET_HTTP_PROXY = <<EOT;
If you use an HTTP proxy server to access the Internet, you need to specify its address to allow the application
to connect to the updates source.
Please enter the address of your HTTP proxy server in one of the following formats: proxyIP:port or user:pass\@proxyIP:port.

If you don\'t have or need a proxy server to access the Internet, enter 'no' here, or enter 'skip' to use current settings without changes.
EOT

my $MSG_SET_PROXY_UPDATED = <<EOT;
Proxy settings is set to '%1'
EOT

my $HEADER_SETUP_AUTOUPDATES = <<EOT;
Enabling scheduled updates of the application databases
EOT

my $MSG_SET_AUTOUPDATES = <<EOT;
Would you like to enable scheduled updates?
EOT

my $MSG_SET_AUTOUPDATES_SUCCESS = <<EOT;
Scheduled updates are enabled successfully.
%1
EOT

my $MSG_SET_AUTOUPDATES_FAILED = <<EOT;
An error occurred while enabling scheduled updates: %1.
EOT

my $HEADER_DOWNLOAD_BASES = <<EOT;
Downloading the latest application databases
EOT

my $MSG_RUN_UPDATER = <<EOT;
The latest databases are an essential part of your server protection.
Would you like to download the latest databases now?
(If you answer 'yes', make sure you are connected to the Internet):
EOT

my $MSG_NOT_CONFIGURED_NOTICE = <<EOT;
<B>%1</B> has been installed successfully, but it must be properly configured before using.
Please run <B>%2</B> script manually to configure it.
EOT

my $MSG_UPGRADE_NOTICE = <<EOT;
<B>%1</B> has been updated successfully, but it must be properly upgraded before using.
Please run <B>%2</B> script manually to perform upgrade.
EOT

my $MSG_AI_EXECUTED_INFO = <<EOT;
Executing automatic installation using parameters from file '%1'
EOT

my $MSG_AI_LOAD_FAILED = <<EOT;
Cannot read parameters of automatic installation from file '%1'
EOT

my $MSG_AI_SAVE_FAILED = <<EOT;
Cannot write automatic installation parameter %1=%2 to file %3: %4
EOT

my $MSG_AI_VALIDATION_FAILED = <<EOT;
Automatic installation failed, invalid values found.
EOT

my $MSG_AI_INSTALLATION_CANCELED = <<EOT;
Automatic installation is canceled.
EOT

my $MSG_CMD_HELP_POSTINSTALL = "--postinstall\tperform post-install configuration (default)";
my $MSG_CMD_HELP_AUTOINSTALL = "--auto-install=<file>\tperform automatic post-install configuration using values from file";
my $MSG_CMD_HELP_HELP = "-h, --help\tshow usage syntax";
my $MSG_CMD_HELP_AI_SAVE = "--create-auto-install=<file>\tsave all answers to file specified";

sub new
{
    my $class = shift;
    my $appid = shift;

    my $self = {
        'SETTINGS' => {},
        'OPTIONS'  => {},
        'DEFAULT_COMMAND' => 'postinstall',
    };

    $self->{'COMMANDS'} = [
        {
            'id' => 'postinstall',
            'spec' => 'postinstall',
            'sub' => 'do_postinstall',
            'descr' => $MSG_CMD_HELP_POSTINSTALL,
        },
        {
            'id' => 'autoinstall',
            'spec' => 'auto-install=s',
            'sub' => 'auto_install',
            'descr' => $MSG_CMD_HELP_AUTOINSTALL,
        },
        {
            'id' => 'trigger',
            'spec' => 'trigger=s',
            'sub' => 'do_trigger',
            'title' => 0,
        },
        {
            'id' => 'upgrade',
            'spec' => 'upgrade',
        },
        {
            'id' => 'debug',
            'spec' => 'd|debug',
        },
        {
            'id' => 'help',
            'spec' => 'help',
            'sub' => 'do_show_help',
            'default' => undef,
            'descr' => $MSG_CMD_HELP_HELP,
        },
        {
            'id' => 'create-auto-install',
            'spec' => 'create-auto-install=s',
            'descr' =>  $MSG_CMD_HELP_AI_SAVE,
        },

    ];

    $self->{'AUTOINSTALL'} = [];

    $self->{APPINFO} = new KL::AppInfo($appid);

    KL::Utils::fatal(2, "Cannot read info file for application ID '%1'", $appid)
        if !defined $self->{APPINFO}->{VALUES};

    bless $self, $class;

    $self->{'SETUP'} = undef;
    $self->{'TITLE'} = $self->get_name_version();

    $self->{'EULA_L10N'} = 0;
    $self->{'EULA_LANG'} = undef;

    $self->{'_MD5_CACHE'} = {};

    return $self;
}

sub appinfo
{
    my $self = shift;

    return $self->{APPINFO}->value(@_);
}

sub get_path
{
    my $self = shift;
    my $name = shift;

    return $self->{APPINFO}->value('paths', $name);
}

sub get_name_version
{
    my $self = shift;

    my $title = $self->appinfo('product', 'title') .
        ' version ' . $self->appinfo('product', 'version') . '/' . $self->appinfo('product', 'release');
}

sub get_applog_path
{
    my $self = shift;
    return $self->get_path('logs') . "/install.log";
}

sub applog
{
    my $self = shift;
    my $message = shift;

    my $logfile = $self->get_applog_path();
    my $stamp = strftime('[%d/%m/%y %H:%M:%S]', localtime);

    my @lines = split /\n/, $message;
    foreach (@lines) {
        $_ = "$stamp $_";
    }

    $message = join("\n", @lines) . "\n";

    if ( open F, ">>$logfile" ) {
                print F $message;
                close F;
    } else {
        print STDERR "LOG: $message";
    }
}

sub options
{
    my $self = shift;
    my $name = shift;
    my $value = shift;

    if (defined $value) {
        $self->{OPTIONS}->{$name} = $value;
    }

    return $self->{OPTIONS}->{$name};
}

sub settings
{
    my $self = shift;
    my $name = shift;
    my $value = shift;

    if (defined $value) {
        $self->{SETTINGS}->{''}->{$name} = $value;
        $self->save_settings();

        KL::Utils::debug("('$name','$value')");
    } else {
        $value = $self->{SETTINGS}->{''}->{$name};
        KL::Utils::debug("('$name') = " . (defined $value ? "'$value'" : 'undef'));
    }

    return $value;
}

sub remove_settings
{
    my $self = shift;

    while (@_) {
        my $name = shift;

        delete $self->{SETTINGS}->{''}->{$name};
    }

    $self->save_settings();
}

sub load_settings
{
    my $self = shift;
    my $filename = $self->get_path('var');

    $self->{'SETTINGS'} = {};

    return if $filename eq '';

    $filename .= '/installer.dat';

    return 1 if ! -f $filename;

    my $data = KL::Utils::load_ini_file($filename);
    if (defined $data) {
        $self->{'SETTINGS'} = $data ;
        return 1;
    }

    return 0;
}

sub save_settings
{
    my $self = shift;
    my $filename = $self->get_path('var');

    return if $filename eq '';

    $filename .= '/installer.dat';

    return KL::Utils::save_ini_file($filename, $self->{'SETTINGS'});
}

sub _md5sum
{
    my $self = shift;
    my $file = shift;

    my $checksum = $self->{'_MD5_CACHE'}->{$file};

    if (defined $checksum) {
        KL::Utils::debug("$checksum $file (cached)");
        return $checksum;
    }

    use Digest::MD5;

    my $digest = Digest::MD5->new;

    open (my $fh, '<', $file) or die "Can't open '$file': $!";
    binmode ($fh);
    $digest->addfile($fh);
    close($fh);

    $checksum = $digest->hexdigest;
    KL::Utils::debug("$checksum $file");

    $self->{'_MD5_CACHE'}->{$file} = $checksum;

    return $checksum;
}

sub is_EULA_agreed
{
    my $self = shift;
    my $agreement_name = shift;
    my $eula_file = shift;

    my $eula_agreed = $self->settings($agreement_name);

    return '' if not defined $eula_agreed;

    if (defined $eula_file) {
        my $current_ver = $self->_md5sum($eula_file);
        KL::Utils::debug("current ver $current_ver");

        my $agreed_ver = $self->settings($agreement_name . '_VER');
        KL::Utils::debug("agreed ver $agreed_ver");

        if ($agreed_ver ne $current_ver) {
            KL::Utils::debug("$eula_file version changed");
            return 'newver';
        }
    }

    return $eula_agreed;
}

sub _check_EULA_l10n
{
    my $self = shift;
    my $name = shift;
    my $filepath = shift;

    return 1;
}

sub _get_EULA_l10n
{
    my $self = shift;
    my $filepath = shift;

    if (not $self->{'EULA_L10N'}) {
        return $filepath;
    }

    my $filename = basename($filepath);
    my $filedir = dirname($filepath);

    my $pattern = "^$filename" . '\.([^.]+)$';

    my $dir;
    opendir ($dir, $filedir) || KL::Utils::fatal(1, $MSG_FAIL_OPENDIR, $filedir);
    my @files = grep { -f "$filedir/$_" } readdir($dir);
    closedir $dir;

    my %licenses = ();
    foreach my $file (@files) {
        chomp($file);
        if ($file =~ /$pattern/) {
            my $lang = $1;
            if ($self->_check_EULA_l10n($lang, $file)) {
                $licenses{$lang} = "$filedir/$file";
            }
        }
    }

    my @langs = sort keys %licenses;

    my $langs_count = scalar(@langs);
    if ($langs_count == 0) {
        KL::Utils::fatal(1, $MSG_NO_LICENSE, $self->appinfo('product', 'title'));
    } elsif ($langs_count == 1) {
        return $licenses{ $langs[0] };
    }

    my $lang = $self->{'EULA_LANG'};
    if ((not defined $lang) or (not defined $licenses{$lang})) {
        my $variants = '';
        my $idx = 1;

        foreach my $lang (@langs) {
            my $title = $lang;
            $title = "$1 ($2)" if $title =~ /^(.*?)-(.*)$/;
            $variants .= "\n$idx) $title";
            $idx++;
        }

        my $choose;
        while (1) {
            $choose = KL::Utils::ask_question('1', $MSG_SELECT_LANG, $variants);
            if ($choose !~ /^(\d+)$/) {
                next;
            }

            if ( ($choose <= 0) || ($choose > scalar @langs) ) {
                next;
            }

            last;
        }
        $choose--;
        $lang = $langs[$choose];

        $self->{'EULA_LANG'} = $lang;
    }

    return $licenses{$lang};
}

sub show_EULA
{
    my $self = shift;

    return $self->show_AGREEMENT(@_);
}

sub show_KSN_EULA
{
    my $self = shift;
    my $eula_file = shift;

    return $self->show_AGREEMENT($eula_file, 'KSN_AGREED', 0, $MSGS_EULA_KSN);
}

sub show_KSN_STATISTICS_EULA
{
    my $self = shift;
    my $eula_file = shift;

    return $self->show_AGREEMENT($eula_file, 'KSN_STATISTICS_AGREED', 0, $MSGS_EULA_KSN_STATISTICS);
}

sub show_PRIVACY_POLICY
{
    my $self = shift;
    my $privacy_policy_file = shift;

    return $self->show_AGREEMENT($privacy_policy_file, 'PRIVACY_POLICY_AGREED', 1, $MSGS_PRIVACY_POLICY);
}

sub decline_KSN_STATISTICS_EULA
{
    my $self = shift;
    my $eula_file = shift;

    KL::Utils::warning($_MSG_KSN_STATISTICS_SHOW_EULA_NOT_ACCEPTED, 0);

    $self->ai_save('KSN_STATISTICS_AGREED', 'no');
    $self->settings('KSN_STATISTICS_AGREED', 'no');
    $self->settings('KSN_STATISTICS_AGREED_VER', $self->_md5sum($eula_file));
    $self->save_settings();

    return 0;
}

sub show_AGREEMENT {
    my $self = shift;
    my $eula_filename = shift;
    my $agreement_name = shift;
    my $must_accept = shift;
    my $msgs = shift;

    $agreement_name = 'EULA_AGREED' unless defined $agreement_name;
    $must_accept = 1 unless defined $must_accept;
    $msgs = $MSGS_EULA unless defined $msgs;

    my $result = 1;

    my $agreed = $self->is_EULA_agreed($agreement_name, $eula_filename);

    if ($agreed eq 'newver') {
        $self->settings($agreement_name, 'no');
    }

    if ($agreed ne 'yes') {
        KL::Utils::header ($msgs->{'header'});

        my $eula_file = $self->_get_EULA_l10n($eula_filename);

        my $pager;
        my $message = $msgs->{'begin'};

        if (defined($ENV{'PAGER'})) {
            $pager = $ENV{'PAGER'};
        } else {
            foreach my $bin ('/usr/bin/less', '/bin/more', '/bin/cat') {
                if ( -x  $bin ) {
                    $pager = $bin;
                    last;
                }
            }
        }

        my $less_charset = "utf-8";
        if ($eula_file =~ /-koi8-?r$/i) {
            $less_charset = 'koi8-r';
        } elsif ($eula_file =~ /-(latin1|iso8859)$/i) {
            $less_charset = 'iso8859';
        } elsif ($eula_file =~ /-(cp1251|windows)$/i) {
            $less_charset = 'windows';
        }

        $ENV{'LESSCHARSET'} = $less_charset;

        if ($pager =~ /less/) {
            $pager .= ' --quit-at-eof';
            $message .= $msgs->{'quit'};
        } elsif ($pager =~ /more/) {
            $message .= $msgs->{'quit'};
        }

        $message .= $msgs->{'end'};

        KL::Utils::ask_question('', $message);

        my $rc = -1;
        if (defined $pager) {
            $rc = system("$pager $eula_file");
        }

        if ($rc != 0) {
            open(EULA, "<$eula_file") or
                KL::Utils::fatal(2, $msgs->{'error'}, $eula_file, $!);

            while (<EULA>) {
                print;
            }
            close(EULA);
        }
        print "\n";

        if (KL::Utils::ask_boolean('',  $msgs->{'question'}) == 0) {
            if ($must_accept) {
                KL::Utils::fatal(2, $msgs->{'declined'}, 0);
            } else {
                KL::Utils::warning($msgs->{'declined'}, 0);
                $self->ai_save($agreement_name, 'no');
                $self->settings($agreement_name, 'no');
                $result = 0;
            }
        } else {
            $self->ai_save($agreement_name, 'yes');
            $self->settings($agreement_name, 'yes');
            $result = 1;
        }

        $self->settings($agreement_name . '_VER', $self->_md5sum($eula_filename));

        $self->save_settings();
    }

    return $result;
}

sub check_key_installed
{
    my $self = shift;

    KL::Utils::warning("App::check_key_installed(...) is not implemented!");

    return (0, "check_key_installed not implemented");
}

sub check_key_file
{
    my $self = shift;
    my $file = shift;

    KL::Utils::warning("App::check_key_file(...) is not implemented!");

    return (0, "check_key_file not implemented");
}

sub install_key_file
{
    my $self = shift;
    my $file = shift;

    KL::Utils::warning("App::install_key_file(...) is not implemented!");

    return (0, "App::install_key_file not implemented");
}

sub get_http_proxy
{
    my $self = shift;

    KL::Utils::warning("App::get_http_proxy(...) is not implemented!");

    return (0, "App::get_http_proxy not implemented");
}

sub set_http_proxy
{
    my $self = shift;
    my $value = shift;

    KL::Utils::warning("App::set_http_proxy(...) is not implemented!");

    return (0, "App::set_http_proxy not implemented");
}

sub run_updater
{
    my $self = shift;

    KL::Utils::warning("App::run_updater(...) is not implemented!");

    return 0;
}

sub enable_autoupdates
{
    my $self = shift;

    KL::Utils::warning("App::enable_autoupdates(...) is not implemented!");

    return (0, "App::enable_autoupdates not implemented");
}

sub do_auto_install
{
    my $self = shift;

    KL::Utils::warning("App::do_auto_install(...) is not implemented!");

    return 0;
}

sub do_postinstall
{
    my $self = shift;

    KL::Utils::warning("App::do_postinstall(...) is not implemented!");

    return 0;
}

sub do_uninstall
{
    my $self = shift;
    my $upgrade = shift;

    KL::Utils::warning("App::do_uninstall(...) is not implemented!");

    return 0;
}


sub save_ai_file
{
    my $self = shift;
    my $filename = shift;
    my $data = shift;

    if ( !open FILE, ">$filename" ) {
        return 0;
    }

    my @names = keys %{$data->{''}};

    foreach my $rule (@{$self->{'AUTOINSTALL'}}) {
        my $name = $rule->{'id'};

        if (defined $data->{''}->{$name}) {
            my $value = $data->{''}->{$name};
            print FILE "${name}=$value\n";
            @names = grep {!/^$name$/} @names;
        }
    }

    print FILE "\n";

    foreach my $name (sort @names) {
        my $value = $data->{''}->{$name};
        print FILE "${name}=$value\n";
    }

    close FILE;

    return 1;
}

sub ai_save
{
    my $self = shift;
    my $name = shift;
    my $value = shift;

    return if not defined($self->{'AI_SAVE_FILE'});

    my $filename = $self->{'AI_SAVE_FILE'};

    my $values = {};
    if ( -f $filename ) {
        $values = KL::Utils::load_ini_file($filename);
    }

    $values->{''}->{$name} = $value;

    if (not $self->save_ai_file($filename, $values)) {
        KL::Utils::warning($MSG_AI_SAVE_FAILED, $name, $value, $filename, $!);
    }
}

sub ai_load
{
    my $self = shift;
    my $filename = shift;

    $self->{AI_VALUES} = {};

    my $values = KL::Utils::load_ini_file($filename);

    if ($values) {
        $self->{AI_VALUES} = $values;
        return 1;
    }

    return 0;
}

sub ai_validate
{
    my $self = shift;

    my $errors = 0;

    foreach my $rule (@{$self->{'AUTOINSTALL'}}) {
        my $id = $rule->{'id'};
        my $type = $rule->{'type'};

        my $value = $self->ai_value($id);
        if (ref($value) eq 'ARRAY') {
            KL::Utils::fatal(1,"redefinition of parameter %1", $id);
        }

        if (defined $value) {
            my ($res, $error);
            if (defined $rule->{'list'}) {
                ($res, $error) = $self->ai_validate_list($rule, $value);
                next if ($res);
            }

            if (defined $type) {
                my $function = "ai_validate_${type}";
                ($res, $error) = $self->$function($rule, $value);
                next if ($res);
            }

            $errors++;

            if (defined $error) {
                KL::Utils::error("'%2' is invalid value for '%1': %3.\n", $id, $value, $error);
            } else {
                KL::Utils::error("'%2' is invalid value for '%1'.\n", $id, $value);
            }
        } else {
            if (defined($rule->{'required'}) and $rule->{'required'}) {
                KL::Utils::error("Required parameter '%1' is not set", $id);
                $errors++;
            }
        }

    }

    return ($errors == 0) ? 1 : 0;
}

sub ai_validate_mask
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    my $id = $rule->{'id'};
    my $mask = $rule->{'mask'};

    if ( $value =~ /$mask/) {
        return 1;
    }

    return (0, undef);
}

sub ai_validate_list
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    my $id = $rule->{'id'};
    my @list = @{$rule->{'list'}};

    foreach my $item (@list) {
        return 1 if $item eq $value;
        $item = "'$item'";
    }

    my $valid_list = pop @list;
    if (scalar(@list)) {
        $valid_list = join(', ', @list) . ' or ' . $valid_list;
    }

    return (0, $valid_list);
}


sub ai_validate_dir
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    if (-d $value) {
        return 1;
    }

    return (0, "is not a path to directory");
}

sub ai_validate_file
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    if (-f $value) {
        return 1;
    }

    return (0, "is not a path to file");
}

sub ai_validate_url
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    my $protocol = '';

    if ($value =~ /^(\w+):\/\/(.*)$/ ) {
        $protocol = $1;
        $value = $2;

        if ( defined $rule->{'protocols'} ) {
            if ( not grep(/^${protocol}$/, @{$rule->{'protocols'}}) ) {
                return (0, "$protocol is not valid protocol");
            }
        }
    }

    if ( ($protocol eq '') and defined($rule->{'protocol_requred'}) and $rule->{'protocol_requred'}) {
        return (0, "protocol must be specified for this URL");
    }

    if ($value =~ /^((.*?)(:(.*?))?@)?(.*?)(:\d+)?/) {
        return 1;
    }

    return (0, "is not valid URL");
}

sub get_default_value
{
    my $self = shift;
    my $name = shift;
    my $result;
    foreach my $rule (@{$self->{'AUTOINSTALL'}}) {
        if ($name eq $rule->{'id'}) {
            if (defined($rule->{'default_value'})) {
                $result = $rule->{'default_value'};
            } else {
                last;
            }
        }
    }
    return $result;
}

sub ai_value
{
    my $self = shift;
    my $section = shift;
    my $name = shift;
    my $result;
    if (!defined ($name)) {
        $name = $section;
        $section = '';
    }
    $result = $self->{AI_VALUES}->{$section}->{$name};
    if (!defined($result)) {
        $result = $self->get_default_value($name);
    }
    return $result;
}

sub _ai_error
{
    my $self = shift;
    my $errorcode = shift;

    if ($errorcode < 70) {
        KL::Utils::fatal($errorcode, $MSG_AI_INSTALLATION_CANCELED);
    }

    if (not defined $self->{_AI_ERROR}) {
        $self->{_AI_ERROR} = $errorcode;
    }
}

sub auto_install
{
    my $self = shift;
    my $values = shift;

    my $file = KL::Utils::detildefy($values->{'autoinstall'});

    KL::Utils::warning($MSG_AI_EXECUTED_INFO, $file);

    if ( !$self->ai_load($file) ) {
        KL::Utils::fatal(2, $MSG_AI_LOAD_FAILED, $file);
    }

    if ($self->ai_validate()) {
        $self->{IS_AUTOINSTALL} = 1;
        $self->{_AI_ERROR} = undef;

        my $res = $self->do_auto_install();

        if (defined $self->{_AI_ERROR}) {
            exit($self->{_AI_ERROR});
        }

        return $res;
    }

    KL::Utils::fatal(2, $MSG_AI_VALIDATION_FAILED);
}


sub trigger_install
{
    my $self = shift;
    my $values = shift;

    my $upgrade = defined $values->{'upgrade'};

    if (defined($ENV{$ENV_AUTO_INSTALL_FILE})) {
        my $file = $ENV{$ENV_AUTO_INSTALL_FILE};

        $self->show_app_title();

        $self->{IS_UPGRADE} = $upgrade;
        $values->{'autoinstall'} = $file;
        KL::Utils::warning($MSG_WARNING_OBSOLETED_ENV_FILE);

        return $self->auto_install($values);
    }

    if ($upgrade && defined $self->{'UPGRADE'}) {
        my $title = $self->{APPINFO}->value('product', 'title');
        KL::Utils::output($MSG_UPGRADE_NOTICE, $title, $self->{'UPGRADE'});
    } else {
        if ( defined $self->{'SETUP'} ) {
            my $title = $self->{APPINFO}->value('product', 'title');
            KL::Utils::output($MSG_NOT_CONFIGURED_NOTICE, $title, $self->{'SETUP'});
        }
    }

    return 1;
}

sub trigger_uninstall
{
    my $self = shift;
    my $values = shift;

    my $upgrade = defined $values->{'upgrade'};

    return $self->do_uninstall($upgrade);
}

sub do_trigger
{
    my $self = shift;
    my $values = shift;

    my $trigger = $values->{'trigger'};

    if ($trigger eq 'install') {
        return $self->trigger_install($values);
    } elsif ($trigger eq 'uninstall') {
        return $self->trigger_uninstall($values);
    }

    KL::Utils::fatal(2, "Value for --trigger must be 'install' or 'uninstall', but not '%1'", $trigger);

    return 0;
}



sub install_key
{
    my $self = shift;

    my $key_file = shift;

    if (defined $key_file) {
        KL::Utils::header ($HEADER_INSTALL_KEY);

        my ($res, $message) = $self->install_key_file( $key_file );
        if ($res) {
            $self->ai_save('INSTALL_KEY', 'yes');
            $self->ai_save('INSTALL_KEY_FILE', $key_file);
            KL::Utils::output ($MSG_INSTALL_KEY_SUCCESS, $key_file);
            return 1;
        }

        KL::Utils::output ($MSG_INSTALL_KEY_FAILED, $key_file, $message);
        return 0;
    }


    my ($res, $message) = $self->check_key_installed();

    return 1 if $res;

    KL::Utils::header ($HEADER_INSTALL_KEY);

    my $keypath = KL::Utils::ask_question ('', $MSG_INSTALL_KEY_ENTER_PATH);
    if ( $keypath ) {

        while ( 1 ) {
            $keypath = KL::Utils::detildefy( $keypath );

            if ( -d $keypath ) {
                my @errormsgs;
                if (!opendir LICDIR, $keypath)
                {
                    push @errormsgs, KL::Utils::translate_string($MSG_INSTALL_OPENDIR_FAILED, $keypath, $!);
                } else {
                    my @keyfiles = grep { /\.key$/i && -f "$keypath/$_" } readdir (LICDIR);
                    closedir LICDIR;

                    if (@keyfiles)
                    {
                        my %validkeys;
                        foreach (@keyfiles)
                        {
                            ($res, $message) = $self->check_key_file("$keypath/$_");
                            if ($res) {
                                $validkeys{"$keypath/$_"} = 1;
                            } else {
                                push @errormsgs, KL::Utils::translate_string($MSG_INSTALL_KEY_REJECTED, $_, $message);
                            }
                        }

                        if ( %validkeys )
                        {
                            @errormsgs = ();
                            my $installed = 0;
                            foreach my $keyfile (keys %validkeys)
                                {
                                $keyfile =~ s/\/+/\//g; # remove extra /
                                ($res, $message) = $self->install_key_file( $keyfile );
                                if ($res) {
                                    $self->ai_save('INSTALL_KEY', 'auto');
                                    $self->ai_save('INSTALL_KEY_FILE', $key_file);
                                    KL::Utils::output ($MSG_INSTALL_KEY_SUCCESS, $keyfile);
                                    return 1;
                                } else {
                                    push @errormsgs, KL::Utils::translate_string($MSG_INSTALL_KEY_FAILED, $keyfile, $message);
                                }
                            }
                        } else {
                            push @errormsgs, $MSG_INSTALL_KEY_NO_VALID_FOUND;
                        }
                    } else {
                        push @errormsgs, KL::Utils::translate_string($MSG_INSTALL_KEY_NO_KEYS_FOUND, $keypath);
                    }
                }

                KL::Utils::output ("\n" . join ("\n", @errormsgs)) if @errormsgs;

            } elsif ( -f $keypath ) {
                ($res, $message) = $self->install_key_file( $keypath );
                if ($res) {
                    $self->ai_save('INSTALL_KEY', 'auto');
                    $self->ai_save('INSTALL_KEY_FILE', $keypath);
                    KL::Utils::output ($MSG_INSTALL_KEY_SUCCESS, $keypath);
                    return 1;
                } else {
                    KL::Utils::output ($MSG_INSTALL_KEY_FAILED, $keypath, $message);
                }
            } else {
                KL::Utils::output ($MSG_INSTALL_KEY_INVALID_PATH);
            }

            $keypath = KL::Utils::ask_question ('', $MSG_INSTALL_KEY_ENTER_OTHER_PATH);
            last if !$keypath;
        }
    }

    return 0;
}

sub setup_proxy
{
    my $self = shift;
    my $value = shift;

    KL::Utils::header ($HEADER_SETUP_PROXY);

    if (defined $value) {
        my ($ok, $str) = $self->set_http_proxy($value);

        if ($ok) {
            KL::Utils::output($MSG_SET_PROXY_UPDATED, $value);
            return 1;
        }

        KL::Utils::error("%1\n\n", $str);
        return 0;
    }

    while ( 1 ) {
        my $default = 'no';

        my ($res, $str) = $self->get_http_proxy();
        if ($res) {
            $default = $str;
        }

        my $value = KL::Utils::ask_question ($default, $MSG_SET_HTTP_PROXY);

        return 0 if $value eq 'skip';

        ($res, $str) = $self->set_http_proxy($value);

        if ($res) {
            $self->ai_save('UPDATER_PROXY', $value);
            return 1;
        }

        KL::Utils::warning("%1\n\n", $str);
    }
}

sub download_bases
{
    my $self = shift;
    my $auto = shift;

    KL::Utils::header($HEADER_DOWNLOAD_BASES);

    if ( !$auto && !KL::Utils::ask_boolean ('yes', $MSG_RUN_UPDATER) ) {
        $self->ai_save('UPDATER_EXECUTE', 'no');
        return;
    }

    $self->ai_save('UPDATER_EXECUTE', 'yes');

    return $self->run_updater();
}

sub setup_autoupdates
{
    my $self = shift;
    my $auto = shift;

    KL::Utils::header ($HEADER_SETUP_AUTOUPDATES);

    if (!$auto && !KL::Utils::ask_boolean('N', $MSG_SET_AUTOUPDATES) ) {
        $self->ai_save('UPDATER_ENABLE_AUTO', 'no');
        return 0;
    }

    $self->ai_save('UPDATER_ENABLE_AUTO', 'yes');

    my ($ok, $message) = $self->enable_autoupdates();
    if (not $ok) {
        KL::Utils::warning($MSG_SET_AUTOUPDATES_FAILED, $message);
        return 0;
    }

    KL::Utils::output($MSG_SET_AUTOUPDATES_SUCCESS, $message);
    return 1;
}

sub show_app_title
{
    my $self = shift;

    KL::Utils::header($self->{'TITLE'});
}

sub do_show_help
{
    my $self = shift;

    print "USAGE: $0 [OPTION]\n";

    my $maxlen = 0;
    my @lines;
    foreach my $option (@{$self->{'COMMANDS'}}) {
        my $descr = $option->{'descr'};

        next if not defined $descr;

        my ($opts, $info) = split(/\t/, $descr, 2);
        push @lines, $opts, $info;

        if ($option->{'id'} eq 'autoinstall') {
            push @lines, "", "Valid auto-install options are:";
            foreach my $auto (@{$self->{'AUTOINSTALL'}}) {
                push @lines, "", "  " . $auto->{'id'} . $auto->{'descr'};
            }
        }

        $maxlen = length($opts) if length($opts) > $maxlen;
    }

    while (@lines) {
        my $opts = shift @lines;
        my $info = shift @lines;

        my $space = $maxlen > length($opts) ? ' ' x ($maxlen - length($opts)) : '';
        print "\t$opts$space\t$info\n";
    }
}

sub parse_cmdline
{
    my $self = shift;

    my %options;
    my @commands;

    my $values = {};
    my %specs;
    foreach my $option (@{$self->{'COMMANDS'}}) {
        my $id = $option->{'id'};
        my $spec = $option->{'spec'};
        my $default = $option->{'default'};

        $values->{$id} = $default;

        $specs{$spec} = \$values->{$id};

        $options{$id} = $option;

        if ($option->{'sub'}) {
            push @commands, $id;
        }
    }

    my $res = GetOptions (%specs);

    if (!$res) {
        $self->do_show_help();
        exit(1);
    }

    KL::Utils::debug_mode(1) if $values->{'debug'};

    my $command;
    foreach my $cmd (@commands) {
        if ( defined $values->{$cmd} ) {
            $command = $cmd;
            last;
        }
    }

    $command = $self->{'DEFAULT_COMMAND'} if not defined $command;

    my $option = $options{$command};

    if ( (not defined $option->{'title'}) or $option->{'title'} ) {
        $self->show_app_title();
    }

    if ( defined($values->{'create-auto-install'}) ) {
        $self->{'AI_SAVE_FILE'} =  KL::Utils::detildefy($values->{'create-auto-install'});
    } elsif ( defined($ENV{$ENV_AUTO_INSTALL_SAVE}) ) {
        $self->{'AI_SAVE_FILE'} = $ENV{$ENV_AUTO_INSTALL_SAVE};
        KL::Utils::warning($MSG_WARNING_OBSOLETED_ENV_SAVEFILE);
    }

    my $function = $option->{'sub'};

    return $self->$function($values);
}

1;
