# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================

package KL::LmsUtils;

use POSIX;
use KL::Utils;
use File::Basename;
use KL::AppBase;

use strict;
use warnings;

sub get_klms_databases_list
{
	return ("configurator", "rule_storage", "backup", "product_status", "notifier", "statistics", "personal_settings",
		"kata_quarantine", "asp_quarantine", "audit_database");
}

sub is_empty
{
    my $value = shift;
    if (!defined($value) || ($value eq "") ) {
        return 1;
    }else{
        return 0;
    }
}

sub not_empty
{
    my $value = shift;
    if ( defined($value) && ($value ne "") ) {
        return 1;
    }else{
        return 0;
    }
}

sub is_not_empty
{
    my $value = shift;
    return not_empty($value);
}

sub trim
{
	my $string = shift;
	if (!defined($string)) {
		return "";
	}
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}

sub print_stack {
	print "Call stack:\n";
	my $i = 1;
	for (;;) {
		my ($package, $filename, $line, $subroutine) = caller($i);
		last if (!defined(scalar caller($i)));
		print "Package $package, sub $subroutine, $filename:$line\n";
		$i++;
	}
}

sub assert_not_empty
{
	my $var = shift;
 	my $name = shift;
	if (is_empty( $var )) {
		print_stack();
		KL::Utils::fatal(1, "ASSERT: Parameter \'$name\' is not set");
	}
}

sub get_temp_file
{
	my $suffix = shift || "lms_utils_pm_temp_file";
	return "/tmp/temp_file_${suffix}_$$";
}

sub get_exit_code
{
	my $procStatus = shift;
	return ($procStatus >> 8);
}

sub run_as
{
    my $user = shift;
    my $cmd = shift;
    my $rundir = shift;

	my $debug = KL::Utils::debug_mode();

	assert_not_empty($user, "user");
	assert_not_empty($cmd, "cmd");

	KL::Utils::debug("Running command \'$cmd\'");

    my ($login,$pass,$uid,$gid,
         $quota,$comment,$gcos,$dir,$shell,$expire) = getpwnam($user)
           or KL::Utils::fatal(1, "$user not exists");

    if (!defined($rundir))
    {
	$rundir = $dir;
    }

    my $pid= fork();
    if( !defined($pid) ) {
        KL::Utils::fatal(2, "Can't fork for command $cmd");
    }

    if($pid == 0)
    {
		POSIX::setgid($gid) or KL::Utils::fatal(2, "Cannot run command '$cmd' as user $user: setgid($gid) failed.");
		POSIX::setuid($uid) or KL::Utils::fatal(2, "Cannot run command '$cmd' as user $user: setuid($uid) failed.");
        chdir($rundir) or KL::Utils::fatal(2, "Cannot run command as user $user: chdir($rundir) failed.");
        my $out =`$cmd 2>&1`;
        my $status = $?;
        if ($status != 0) {
            print "$out\n" if $out ne '';
        } else {
            KL::Utils::debug("out: $out") if $out ne '';
        }
        exit get_exit_code($status);
    } else {
        my $wpid = wait();
        my $status = $?;
        return get_exit_code($status);
    }
}


my $MSG_ERR_CREATE_PATH=<<EOT;
Can not create path %1
EOT
my $MSG_ERR_CHOWN_PATH=<<EOT;
Can not change owner of %1 to %2
EOT
my $MSG_ERR_BAD_PATH=<<EOT;
Path %1 can not be created. There is file (or symlink, etc.) with specified name.
EOT
my $MSG_ERR_BAD_PATH_PERM=<<EOT;
Path %1 is not accessible for %2.
EOT


sub remove_dir
{
    my $path = shift;
    assert_not_empty($path, "path");

	use File::Path;
	rmtree($path);
}

sub create_dir
{
    my $new_path = shift;

    assert_not_empty($new_path, "new_path");

    my $userName = shift || 'kluser';
    my $mode = shift;

    use File::Spec;
    $new_path = File::Spec->rel2abs( $new_path ) ;
    if (! -e $new_path) {
        use File::Path;
        eval { mkpath($new_path); };

        if ($@) {
             KL::Utils::error( $MSG_ERR_CREATE_PATH, $new_path);
             return 1;
        }
    }

    if ( -d $new_path){
        my ($login,$pass,$uid,$gid) = getpwnam($userName)
               or KL::Utils::fatal(1, "$userName not exists");
        my $changed = chown $uid, $gid, $new_path;
        if (defined($mode) ) {
            chmod $mode, $new_path;
        }
        if ( $changed !=1 ) {
            KL::Utils::error( $MSG_ERR_CHOWN_PATH, $new_path);
            return 3;
        }
        local $>=$uid;
        local $)=$gid;
        if ( -x $new_path && -w $new_path && -r $new_path){
            return 0;
        }else{
            KL::Utils::error( $MSG_ERR_BAD_PATH_PERM, $new_path, $userName);
            return 4;
        }
    } else {
        KL::Utils::error( $MSG_ERR_BAD_PATH, $new_path, $userName);
        return 5;
    }
    return 6;
}


my $MSG_INIT_SCRIPT_NOT_FOUND=<<EOT;
Init script not found for %1 daemon.
EOT

sub find_service
{
	my @serviceList = @_;

	for my $serviceName (@serviceList) {
		if (is_not_empty($serviceName)) {
			my $serviceInfo = find_one_service($serviceName);
			if ($serviceInfo) {
				KL::Utils::debug("Found service by name $serviceName");
				return $serviceInfo;
			}
		}
	}
	return undef;
}

sub is_rc_script
{
	my $script = shift;
	if (is_empty($script)) {
		return 0;
	}
	if ($script =~ /^\/\w+/) { # /word
		return 1;
	}
	return 0;
}

sub is_absolute_path
{
	my $path = shift;
	if (is_empty($path)) {
		return 0;
	}
	if ($path =~ /^\//) { # /...
		return 1;
	}
	return 0;
}

sub find_one_service
{
	my $serviceName = shift;
    assert_not_empty($serviceName, "serviceName");

    if (is_rc_script($serviceName)) {
		if (-f $serviceName) {
			return {
				"name" => File::Basename::basename( $serviceName),
				"type" => "SysV",
				"rc" => $serviceName
			}
		} else {
			return undef;
		}
    }

    my $sysctl = KL::Utils::which("systemctl");
    if (is_not_empty($sysctl)) {
		KL::Utils::debug("Using systemctl");
		return find_one_service_using_sysctl($serviceName);
    }
    KL::Utils::debug("Using rc-scripts");

    return find_one_service_using_rcd($serviceName);
}

sub find_one_service_using_sysctl
{
	my $serviceName = shift;
	assert_not_empty($serviceName, "serviceName");
    my $sysctl = KL::Utils::which("systemctl");
    if (is_empty($sysctl)) {
		KL::Utils::fatal(1, "Bug in find_service logic: systemctl not found, but we tried to use 'find_one_service_using_sysctl' function. ");
    }
    my ($ret, $output) = KL::Utils::execute("$sysctl", "--no-pager", "--type=service", "show", "${serviceName}.service", "--property=LoadState");
    if ($ret == 0) {
		KL::Utils::debug("sysctl output: $output");
		if ($output =~ /LoadState=loaded/) {
			return {
				"name" => $serviceName,
				"type" => "systemd",
				"rc" => ""
			};
		}
    } else {
		KL::Utils::debug("systemctl returned $ret, out: $output");
    }
    return undef;
}

sub find_one_service_using_rcd
{
	my $serviceName = shift;
	assert_not_empty($serviceName, "serviceName");

	my $rc = "/etc/init.d/$serviceName";
	if (KL::Utils::is_freebsd()) {
		$rc = "/usr/local/etc/rc.d/$serviceName";
		if (! -e $rc ) {
			$rc = "$rc.sh";
			if (! -e $rc ) {
				$rc = "/etc/rc.d/$serviceName";
				if (! -e $rc ) {
					$rc = "$rc.sh";
				}
			}
		}
	}
	KL::Utils::debug("rc-script: $rc");

	if (! -e $rc) {
		return undef;
	}

	KL::Utils::debug("Found rc-script $rc for daemon $serviceName");

	return {
		"name" => $serviceName,
		"rc" => $rc,
		"type" => "SysV"
		};
}

sub service
{
    my $name = shift;
    my $action = shift;
    my $terminateOnError = shift;

    assert_not_empty($name, "name");
    assert_not_empty($action, "action");

    my $serviceInfo = find_service($name);
    if (!$serviceInfo) {
        KL::Utils::error("Service $name not found");
        if ($terminateOnError) {
            KL::Utils::fatal(1, KL::LmsMsg->MSG_SERVICE_OPERATION_FAILED, $name);
        }
        return 2;
    }

    my $serviceType = $serviceInfo->{'type'};
    my $serviceName = $serviceInfo->{'name'};

    if ($serviceType eq "systemd") {
        KL::Utils::debug("Controlling service $serviceName using systemd. ")
    } else {
        KL::Utils::debug("Controlling service $serviceName using SysV. ")
    }

    if (KL::Utils::is_freebsd()) {
        $action = "one$action";
    }

    print "Executing action '$action' on the service '$serviceName':\n" if $action !~ /^(one)?status$/;

    my @cmd;
    my $cmd_name;

    my $serviceCmd = KL::Utils::which("service");
    if ( is_not_empty($serviceCmd) ) {
        $cmd_name = 'Service';
        @cmd = ($serviceCmd, $serviceName, $action);
    } else {
        if ($serviceType eq "systemd") {
            KL::Utils::fatal(1, "Cannot find 'service' command, but service $serviceName is controlled by systemd. ");
        }

        my $rc = $serviceInfo->{'rc'};

        KL::Utils::debug("Using rc-script $rc for daemon $serviceName");

        my $invoke_rc_d = KL::Utils::which("invoke-rc.d");
        if (is_not_empty( $invoke_rc_d)) {
            $cmd_name = 'invoke-rc.d';
            @cmd = ($invoke_rc_d, $serviceName, $action);
        } else {
            $cmd_name = 'rc-script';
            @cmd = ($rc, $action);
        }
    }

    if ($action =~ /^(one)?status$/) {
        my ($ret, $out) = KL::Utils::execute(@cmd);
        KL::Utils::debug("$cmd_name execution result: $ret, output: $out");
        return $ret;
    } else {
        my $ret = KL::Utils::system(@cmd);
        KL::Utils::debug("$cmd_name command result: $ret");
        return $ret;
    }
}

sub service_is_running
{
    my $name = shift;
    my $terminateOnError = shift;

    my $res = service($name, 'status', $terminateOnError);
    if ($res == 0) {
        return 1;
    }
    return 0;
}

sub service_is_not_running
{
    my $name = shift;
    my $terminateOnError = shift;
    return (! service_is_running($name, $terminateOnError));
}

sub service_start
{
    my $name = shift;
    my $terminateOnError = shift;
    return service($name, 'start', $terminateOnError);
}

sub service_stop
{
    my $name = shift;
    my $terminateOnError = shift;
    return service($name, 'stop', $terminateOnError);
}

sub service_restart
{
    my $name = shift;
    my $terminateOnError = shift;
    return service($name, 'restart', $terminateOnError);
}

sub write_file_atomic
{
	my $file = shift;
	my $content = shift;

	my $temp_file = "$file.tmp";

	if (!KL::Utils::write_file($temp_file, $content)) {
		return 1;
	}
	if (system("mv -f '$temp_file' '$file'") != 0) {
		KL::Utils::error("Cannot move temporary file $temp_file to $file");
		return 2;
	}
	return 0;
}

sub print_string_to_file
{
	my $fileName = shift;
	my $data = shift;

	assert_not_empty($fileName, "fileName");

	my $fd;
	if (!open($fd, ">", $fileName))
	{
		KL::Utils::error("print_string_to_file(): Cannot open file $fileName for writing. ");
		return 1;
	}
	print $fd $data;
	close($fd);
	return 0;
}

sub read_string_from_file
{
	my $fileName = shift;
	assert_not_empty($fileName, "fileName");

	my $fd;
	if (!open($fd, "<", $fileName))
	{
		KL::Utils::error("read_string_from_file(): Cannot open file $fileName for reading. ");
		return undef;
	}
	local $/ = undef;
	my $content = <$fd>;
	close($fd);
	return $content;
}

sub get_user
{
	my $appBase = shift;
	my $user = $appBase->appinfo('credentials', 'user');
	if (is_empty($user))
	{
		KL::Utils::error("Variable credentials::user is empty");
	}
	return $user;
}

sub get_group
{
    my $appBase = shift;
    my $group = $appBase->appinfo('credentials', 'group');
    if (is_empty($group))
    {
        KL::Utils::error("Variable credentials::group is empty");
    }
    return $group;
}

sub add_user_to_group
{
    my $username = shift;
    my $groupname = shift;

    assert_not_empty($username, "username");
    assert_not_empty($groupname, "groupname");

    my $error_if_group_not_exists = shift;
    $error_if_group_not_exists = 1 unless defined($error_if_group_not_exists);

    my $gid=getgrnam($groupname);
    if(defined($gid) && ( $gid=~/^\d+$/)) {
        KL::Utils::output("Adding $username to group $groupname (gid = $gid)");
        if ( KL::Utils::is_freebsd() ) {
            `pw groupmod -g $gid -m $username`;
        }else{
            `usermod  -G $gid $username`;
        }
        if ( $? != 0) {
            KL::Utils::fatal(1,"Can't add '$username' to group '$gid'");
        }else{
            KL::Utils::output("Success");
            return 1;
        }
    } elsif ($error_if_group_not_exists) {
        KL::Utils::fatal(1,"Group '$groupname' not exists");

    }else{
        return 0;
    }
}
sub file_replace_str {
    my  $file = shift;

    assert_not_empty($file, "file");

    my  $regex= shift;
    my  $dest = shift;
    my $fd_in;
    my $fd_out;
    if( !open ($fd_in, "<", $file)) {
        die "Internal error: can't open $file:$!";
    }
    if( !open ($fd_out, ">", "$file.tmp")) {
        die "Internal error: can't open $file.tmp:$!";
    }
    while(<$fd_in>) {
        my $line = $_;
        chomp($line);
        $line =~ s/$regex/$dest/g;
        print $fd_out "$line\n"
    }
    close $fd_in;
    close $fd_out;
    system "mv -f $file.tmp $file";
}

sub read_password_from_console
{
    my $prompt = shift;
    if (defined($prompt)) {
        KL::Utils::output($prompt,@_);
    }
    system("stty -echo");
    my $password = <>;
    system("stty echo");
    $password = trim($password);
    return $password;
}

sub get_my_pids
{
    my $scriptName = shift;
    assert_not_empty($scriptName, "scriptName");

    my $pids=();
    for my $line (`ps ax|grep perl|grep $scriptName`) {
        chomp $line;
        if($line =~ /^\s*(\d+)\s+\S+\s+\S+\s+\S+\s+(.*)$/) {
                $pids->{$1}=$2;
        }
    }
    return $pids;
}

sub setup_is_already_running
{
    my $pid_file = shift;
    my $my_name = shift;

    assert_not_empty($pid_file, "pid_file");
    assert_not_empty($my_name, "my_name");

    if (! -f $pid_file ) {
        return 0;
    }
    my $fd;
    if (!open $fd,"<",$pid_file) {
        KL::Utils::fatal(1,"Can't open $pid_file:$!");
    }
    my $pid =<$fd>;
    chomp $pid;
    close $fd;
    if ( $pid == $$ ) {
        return 0;
    }else{
        my $pids = &get_my_pids($my_name);
        if (!exists($pids->{$pid})) {
            unlink $pid_file;
            return 0;
        }else{
            return 1;
        }
    }
}
sub create_pid_file
{
    my $pid_file = shift;
    assert_not_empty($pid_file, "pid_file");

    my $fd;
    if (!open $fd,">",$pid_file) {
        KL::Utils::fatal(1,"Can't open $pid_file:$!");
    }
    print $fd  "$$";
    close $fd;
}

sub set_utf8_locale
{
	use POSIX qw(locale_h);

	my $locOutput = `locale -a`;

	my @locList = split(/\n/, $locOutput);

	my @locFiltered = grep(/(utf8|UTF-8)$/, @locList);

	my @locEng = grep(/^en_US.(utf8|UTF-8)$/, @locFiltered);

	my $targetLocale = "";

	if (@locEng >= 1) {
		$targetLocale = $locEng[0];
	} else {
		$targetLocale = $locFiltered[0];
	}
	if ($targetLocale eq "") {
		KL::Utils::error("Cannot find UTF locale. ");
		return 1;
	}
	$ENV{'LC_ALL'} = "";
	$ENV{'LANG'} = $targetLocale;
	if (!defined(setlocale(LC_ALL, $targetLocale)))
	{
		KL::Utils::error("Cannot set locale to $targetLocale. ");
		return 2;
	}
	return 0;
}

sub install_service
{
    my $name = shift;
    my $priority_start = shift;
    my $priority_stop  = shift;
    my $package = shift;
    my $exit_code = 0;

    KL::Utils::output("Registering service $name");

    if ( $package eq 'deb' ) {
        $exit_code = system "update-rc.d $name start $priority_start 2 3 4 5  . stop $priority_stop 0 1 6 . >/dev/null";
    } elsif ( $package eq 'rpm' ) {
        $exit_code = system "chkconfig --add \"$name\"";
    } else {
my $MSG_BSD_ENABLE_SERVICE = <<EOT;
To enable boot time and manual start of $name service add the following line to /etc/rc.conf:
${name}_enable="YES"
Then you can start $name service using:
/usr/local/etc/rc.d/$name start
EOT
;
        KL::Utils::output($MSG_BSD_ENABLE_SERVICE);
        return 1;
    }

    if ($exit_code != 0) {
        KL::Utils::error(KL::LmsMsg->MSG_FAIL_REGISTER_SERVICE, $name);
        return 0;
    }

    return 1;
}

sub uninstall_service
{
    my $name = shift;
    my $package = shift;
    my $exit_code = 0;

    KL::Utils::output("Unregistering service $name");

    if ( $package eq 'deb' ) {
        $exit_code = system "update-rc.d -f $name remove >/dev/null";
    } elsif ( $package eq 'rpm' ) {
        $exit_code = system "chkconfig --del $name";
    } else {
    }

    if ($exit_code != 0) {
        KL::Utils::error(KL::LmsMsg->MSG_FAIL_UNREGISTER_SERVICE, $name);
        return 0;
    }

    return 1;
}

sub get_hostname
{
    my $default = shift;

    my ($ret, $hostname) = KL::Utils::execute('/bin/hostname', '--fqdn');
    chomp $hostname;
    return lc($hostname) if $ret == 0;

    return $default;
}

sub _is_sha256_supported
{
    my @cmd = ('/usr/bin/openssl', 'dgst', '-sha256', '/dev/null');
    my ($rc, $out) = KL::Utils::execute(@cmd);
    return $rc == 0;
}

sub create_ssl_cert
{
    my $cert = shift;

    my $domain = defined $cert->{'domain'} ? $cert->{'domain'} : get_hostname('localhost');
    my $days = defined $cert->{'days'} ? $cert->{'days'} : 3650;
    my $key = $cert->{'key'};
    my $crt = $cert->{'crt'};

    unlink $key if -f $key;
    unlink $crt if -f $crt;

    my @digest = _is_sha256_supported() ? ('-sha256') : ();

    my @cmd = ('/usr/bin/openssl',
        'req', '-batch', '-x509', @digest,
        '-newkey', 'rsa:2048', '-nodes',
        '-days', $cert->{'days'},
        '-keyout', $key,
        '-out', $crt,
        '-subj', '/CN='.$domain,
    );
    my ($rc, $out) = KL::Utils::execute(@cmd);
    if (($rc != 0) or (! -f $key) or (! -f $crt)) {
        KL::Utils::error("Could not execute the command ".join(' ', @cmd)."\nexecution result: $rc\noutput: $out\n");
        return 0;
    }

    return 1;
}

1;
