# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================

package KL::Utils;

use strict;
use Data::Dumper;

my $DEBUG_MODE = defined($ENV{'KLI_DEBUG_ENABLED'}) && $ENV{'KLI_DEBUG_ENABLED'} eq 'yes' ? 1 : 0;

my $BOLD = "";
my $NORMAL = "";
my $UNDERLINE = "";

if ( -t STDOUT and not defined $ENV{'ANSI_COLORS_DISABLED'} ) {
    $BOLD = "\e\[1m";
    $NORMAL = "\e[0m";
    $UNDERLINE = "\e[4m";
}

sub is_freebsd
{
    return ($^O =~ /^freebsd$/i) ? 1 : 0;
}

sub is_linux
{
    return ($^O =~ /^linux$/i) ? 1 : 0;
}

sub debug_mode
{
    my $mode = shift;

    $DEBUG_MODE = $mode if defined $mode;

    return $DEBUG_MODE;
}

sub get_os_name
{
    my $name = "unknown";

    if ( $^O =~ /linux/i )  {
        if ( -f '/etc/redhat-release' ) {
            $name = read_file('/etc/redhat-release');
        } elsif ( -f '/etc/SuSE-release' ) {
            $name = read_file('/etc/SuSE-release');
        } elsif ( -f '/etc/debian_version' ) {
            $name = "Debian " . read_file('/etc/debian_version');
        }
    } elsif ( $^O =~ /sunos/i ) {
        $name = `uname -prs`;
    } elsif ( $^O =~ /[hp-ux,bsd]/i ) {
        $name = `uname -rs`;
    }

    $name =~ s/\n.*//g;
    chomp $name;
    return $name;
}

sub output
{
    my $text = translate_string(@_) . "\n";

    my $margin = 75 - 1;
    my ($para, $accum, $outstr);

    foreach $para (split ("\n", $text))
    {
        $accum = '';

        my $length = 0;
        my $line = 0;
        my $indent = '';
        if ($para =~ /^<LI>/) {
            $indent = '  ';
            $para =~ s/<\/?LI>//g;
        }

        foreach my $word (split(/\s+/, $para))
        {
            my $noesc = $word;

            if ($BOLD ne '') {
                $noesc =~ s/\Q$BOLD\E|\Q$NORMAL\E|\Q$UNDERLINE\E|<BR>//g;
            }

            if (($word =~ /^<BR>/) or ($length + length($noesc) > $margin)) {
                $accum .= "\n";
                $length = 0;
                $line++;
            }

            $word =~ s/<BR>//g;

            if ($length > 0) {
                $accum .= ' ' . $word;
                $length += length($noesc) + 1;
            } else {
                if ($line > 0) {
                    $accum .= $indent;
                    $length += length($indent);
                }
                $accum .= $word;
                $length += length($noesc);
            }
        }

        $outstr .= $accum . "\n";
    }

    print $outstr if defined $outstr;
    return 1;
}

sub header
{
    my $title = shift;
    chomp $title;

    print "\n";
    if ( $BOLD ne '' ) {
        output("<B><U>%1</B>\n", $title);
    } else {
        output("%1\n%2\n", $title, '=' x length($title));
    }
    print "\n";
}

sub debug
{
    return if not $DEBUG_MODE;

    my $msg = shift;

    my $where = (caller(1))[3];
    print translate_string("<B>DEBUG:</B> ") . "$where: $msg\n";
    print Dumper(@_) if @_;
}

sub fatal
{
    my $exit_code = shift;

    output "\n<B>Fatal error:</B> " . translate_string(@_) . "\n";

    exit $exit_code;
}

sub warning
{
    output "\n<B>Warning:</B> " . translate_string(@_) . "\n";
}

sub error
{
    output "\n<B>Error:</B> " . translate_string(@_) . "\n";
}

sub translate_string
{
    my ($str, @args) = @_;

    $str =~ s/<B>/$BOLD/g;
    $str =~ s/<U>/$UNDERLINE/g;
    $str =~ s/<\/(B|U)>/$NORMAL/g;

    for ( my $i = 0; $i < scalar @args; $i++ )
    {
        warn ("Empty argument $i in string '$str'\n") if !defined $args[$i];
        my $pattern = "%" . ($i + 1);
        $str =~ s/$pattern/$args[$i]/g;
    }

    return $str;
}

sub load_ini_file
{
    my $filename = shift;
    my $data = {};

    return if ! -f $filename;

    open F, "<$filename" or return;
    my $content = join ("", <F>);
    close F;
    $content =~ s/\r//g;

    my $section = '';

    foreach (split (/\n/, $content)) {
        if (/^\s*\[(.*)\]+\s*$/) {
            $section = $1;
            next;
        }

        next if /^\s*[#;].*?$/;
        next if !/^(.*?)=(.*)$/;

        $data->{$section} = {} if !defined $data->{$section};

        if ( !defined $data->{$section}{$1} ) {
            $data->{$section}{$1} = $2;
            next;
        }

        if ( ref $data->{$section}{$1} ne "ARRAY" ) {
            my @arr = ($data->{$section}{$1}, $2);
            $data->{$section}{$1} = \@arr;
        } else {
            push @{$data->{$section}{$1}}, $2;
        }
    }

    return $data;
}

sub save_ini_file
{
    my $filename = shift;
    my $data = shift;

    if ( !open FILE, ">$filename" ) {
        return;
    }

    foreach my $section (sort keys %{$data}) {
        print FILE "[$section]\n" if $section ne '';

        foreach my $name (sort keys %{$data->{$section}}) {
            my $value = $data->{$section}->{$name};
            print FILE "${name}=$value\n";
        }

        print FILE "\n";
    }

    close FILE;
    return 1;
}

sub ask_question
{
    my ($def_answer, $question, @args) = @_;

    my $qtext = translate_string ($question, @args);
    chomp ($qtext);

    output ($qtext . ($def_answer ? " [$def_answer]" : '') . ':');

    my $answer = <STDIN>;
    $answer = "" if !defined $answer;
    $answer =~ s/[\r\n]//g;
    $answer = $def_answer if length($answer) == 0;

    print "\n";
    return $answer;
}


sub ask_list
{
    my ($def_answer, $valud_answ, $question, @args) = @_;
    while ( 1 ) {
        my $tmp = ask_question ($def_answer, $question, @args);
        $tmp =~ s/^\s+(.*?)\s+$/$1/;
        my @answ = split(/\s+/, $tmp);
        my %t0=(); my %t1=(); my %t2=();
        for (@$valud_answ, @answ) { $t0{$_}++ && $t1{$_}++ }
        @t2{@answ} = 1;
        delete @t2{@$valud_answ};
        return keys %t1 if !keys %t2;
    }
}


sub ask_boolean
{
    while ( 1 ) {
        my $answ = ask_question (@_);

        $answ = "yes" if $answ =~ /^yes$/i || $answ =~ /^y$/i;
        $answ = "no"  if $answ =~ /^no$/i  || $answ =~ /^n$/i;

        return 1 if $answ eq "yes";
        return 0 if $answ eq "no";

        output "Please answer either 'yes' or 'no'.\n";
    }
}

sub shell_escape
{
    my $str = shift;
    $str =~ s/([\&;\`'\\\| "*?~<>^\(\)\[\]\{\}\$\n\r])/\\$1/g; #'
    return $str;
}

sub execute
{
    my @args = @_;

    foreach (@args) {
        $_ = shell_escape($_);
    }

    my $cmd = join(' ', @args);

    debug($cmd) if $DEBUG_MODE;

    open(FILE,"$cmd 2>&1|")
        or return -128;

    my ($ret, $output);
    $output = join('', <FILE>);
    close(FILE);
    $ret = $? >> 8;

    return ($ret, $output);
}

sub system
{
    my @args = @_;

    foreach (@args) {
        $_ = shell_escape($_);
    }

    my $cmd = join(' ', @args);

    debug($cmd) if $DEBUG_MODE;

    system(@args) == 0
        or return -128;

    my $ret = $? >> 8;

    return $ret;
}

sub copy_file
{
    my ($src, $dest) = @_;

    $dest .= "/$1" if -d $dest && $src =~ /([^\/]+)$/;

    open FSRC, "<$src" or return "Could not read file $src: $!";
    open FDST, ">$dest" or return "Could not write file $dest: $!";

    binmode FSRC;
    binmode FDST;
    my $length;

    while ( ($length = sysread (FSRC, my $data, 8192)) > 0 )
    {
        if ( syswrite (FDST, $data, $length) != $length )
        {
            close FSRC;
            close FDST;

            unlink $dest;
            return "Could not write file $dest: $!";
        }
    }

    close FSRC;
    close FDST;

    if ( !defined $length )
    {
        unlink $dest;
        return "Couldn't read file $src: $!";
    }

    my @statinfo = stat ($src);
    if ( @statinfo )
    {
        chown $statinfo[4], $statinfo[5], $dest;
        chmod (($statinfo[2] & 07777), $dest);
    }

    return;
}


sub copy_files
{
        my ($srcdir, $destdir, $pattern) = @_;
        my @errormessages;

        if ( opendir (DIR, $srcdir) )
        {
                my @files = grep { -f "$srcdir/$_" && /$pattern/i } readdir DIR;
                closedir DIR;

                foreach (@files)
                {
                        my $msg = klinstall::copy_file ("$srcdir/$_", "$destdir/$_");
                        push @errormessages, $msg if $msg;
                }
        }
        else
        {
                push @errormessages, "Could not open directory $srcdir: $!";
        }

        return @errormessages;
}

sub which
{
    my $cmd = shift;

    foreach my $path (split/:/, $ENV{'PATH'}) {
        return "$path/$cmd" if -x "$path/$cmd";
    }
}

sub read_file
{
    my $filename = shift;

    open(FILE,"<$filename")
        or return;

    my $lines = join ("", <FILE>);
    close(FILE);

    return $lines;
}

sub write_file
{
    my $filename = shift;
    my $lines = shift;

    open F, ">$filename"
        or return;

    my $result = print F $lines;
    close F;

    return $result;
}

sub detildefy
{
    my $path = shift;

    if ($path =~ /^~(\w*)(.*)$/ ) {
        my $home = $1 eq '' ? (getpwuid $>)[7] : (getpwnam $1)[7];
        $path = $home . $2 if $home;
    }

    return $path;
}

1;
